/**
 * @file dmxplayer.c
 * @author Florian Schmitt <florian.schmitt@qikgrp.com>
 * ----------------------------------------------------------------------------
 * (c) Florian Schmitt / 2012-2013
 * License GPLv3+: GNU GPL Version 3 
 * or higher <http://gnu.org/licenses/gpl.html>
 * This code comes with ABSOLUTELY NO WARRANTY. This is free software, and 
 * you are welcome to redistribute it under certain conditions.
 * ----------------------------------------------------------------------------
 * To play and record dmx programs
**/
#include "dmxplayer.h"
#include "dmxset.h"
#include "frametimer.h"

/* create programs */
dmxprogram_t *programs[DMXPLAYER_PROGRAMS];
/* create overwrite cue (direct output) */
dmxlist_t *programmer;
/* socket */
int sock;
/* length of one step */
int tapstep;
long steplen;
long timetap[4];
long lasttap;
/* count quarters for useable time info */
long quarter_cnt;


/**
 * dmxplayer_init()
 * Initiates a dmx player
 * @return 1 on success, 0 on failure
**/
int dmxplayer_init()
{
	int n;
	tapstep = 0;
	lasttap = get_time_usecs();
	for(n=0; n<4; n++) timetap[n] = 1000*1000;
	/* initialize all programs */
	for(n=0; n<DMXPLAYER_PROGRAMS; n++)
	{
		programs[n] = dmxprogram_create();
		if(programs[n] == 0) return 0;
	}
	/* create overwrite scene */
	programmer = dmxcue_init();
	if(programmer==0) return 0;
	/* start frame timer */
	steplen = DMXPLAYER_FPS * 2;
	frametimer_start(&dmxplayer_render, DMXPLAYER_FPS);
	/* success */
	return 1;
}



/**
 * dmxplayer_destroy()
 * Destructor for dmx player
**/
void dmxplayer_destroy()
{
	int n;
	frametimer_stop();
	for(n=0; n<DMXPLAYER_PROGRAMS; n++)
		if(programs[n]!=0) dmxprogram_destroy(programs[n]);
	if(programmer!=0)
		dmxcue_destroy(programmer);
}



/**
 * dmxplayer_render()
 * Render a current frame
 * NOTICE: THIS FUNCTION IS CALLED BY A THREAD TIMER
 *         STARTED BY DMXPLAYER_INIT DONT CALL IT YOURSELF!
 * @param frame Current timecode / Frame number
**/
void dmxplayer_render(long frame)
{
	int n, fadepos;
	/* handle quarter counter */
	if(frame == 0) quarter_cnt = 0;
	if( ( frame % (steplen/4)) == 0 ) quarter_cnt++;
	/* calculate programs */
	for(n=0; n<DMXPLAYER_PROGRAMS; n++)
	{
		int totalduration, fadelength;
		/* calculate step duration of a program (in frames) */
		if(programs[n]->steplen >= 0) 
			totalduration = steplen * (programs[n]->steplen);
		else 
			totalduration = (-1)*steplen / (programs[n]->steplen);
		/* if program wants to be refreshed, do it. */
		if(programs[n]->active && (programs[n]->size > 0) && (programs[n]->refresh > 0))
		{
			/* fadelength: fade-in length in frames */
			fadelength = ( (programs[n]->fade * totalduration) / 100 );
			if((fadelength == 0) || (totalduration==0))
			{
				dmxset_global_loadcue( dmxprogram_get_current_cue(programs[n]), 
					                   programmer );
			}
			else
			{
				/* fader position = (current step duration), but max fadelength */
				fadepos = (frame % totalduration)+1;
				if(fadepos > fadelength) fadepos = fadelength;
				/* promillize value */
				fadepos = (fadepos * 1000) / fadelength;
				dmxset_global_loadcue_faded( dmxprogram_get_current_cue(programs[n]), 
					                         programmer, fadepos );
			}
			programs[n]->refresh = 0;
		}
		/* do stepups for auto trigger programs */
		if( (programs[n]->active) && 
		    (programs[n]->size > 0) && 
		    (totalduration != 0) && 
		    (programs[n]->trigger == 1) )
		{
			if( (frame % totalduration) == 0 ) 
			{
				programs[n]->refresh = 1;
				dmxprogram_step_up(programs[n]);
			}
			/* if fade active refresh every frame */
			else if(programs[n]->fade > 0)
			{
				programs[n]->refresh = 1;
			}
		}
	}
	/* calculate programmer memory - only 1/6 second */
	if( ( frame % (DMXPLAYER_FPS/6)) == 0 ) 
		dmxset_global_loadcue(programmer, 0);
}



/**
 * dmxplayer_set_prog()
 * Sets channel from programmer
 * @param channel   Programmer channel to set
 * @param value     Value to set to
 * @return See: dmxcue_set()
**/
int dmxplayer_set_prog(int channel, int value)
{
	int ret;	
	if(value < 0) value = 0;
	else if(value > 255) value = 255;
	/* Set */
	frametimer_lock();
		 ret = dmxcue_set(channel, value, programmer);
	frametimer_unlock();
	return ret;
}



/**
 * dmxplayer_set_prog_add()
 * Sets channel from programmer additive
 * @param channel   Programmer channel to set
 * @param value     Value to add
 * @return See: dmxcue_set()
**/
int dmxplayer_set_prog_add(int channel, int value)
{
	int newval = dmxcue_get_channel_value(channel, programmer) + value;
	return dmxplayer_set_prog(channel, newval);
}



/**
 * dmxplayer_unset_prog()
 * Removes channel from the programmer
 * @param channel   Programmer channel to unset
 * @return See: dmxcue_unset()
**/
int dmxplayer_unset_prog(int channel)
{
	/* Set */
	int ret;	
	frametimer_lock();
		ret = dmxcue_unset(channel, programmer);
	frametimer_unlock();
	return ret;
}



/**
 * dmxplayer_clear_prog()
 * Reset the whole programmer
 * @return See: dmxcue_clear()
**/
int dmxplayer_clear_prog()
{
	int ret;	
	frametimer_lock();
		ret = dmxcue_clear(programmer);
	frametimer_unlock();
	return ret;
}



/**
 * dmxplayer_is_set_prog()
 * @return 1 if one or more values aer set in programmer
**/
int dmxplayer_is_set_prog()
{
	return !( dmxcue_is_empty(programmer) );
}



/**
 * dmxplayer_get_prog_value()
 * Get a channels value from programmer if existing
 * @param channel   Programmer channel to set
 * @return value or -1 if channel not set
**/
int dmxplayer_get_prog_value(int channel)
{
	/* Set */
	dmxpair_t *p = dmxcue_get_channel_obj(channel, programmer);
	if(p == 0) return -1;
	return p->value;
}



/**
 * dmxplayer_store_progam()
 * Stores programmer into an program
 * @param pid   Program id
 * @return 1 on success
**/
int dmxplayer_store_program(int pid)
{
	dmxlist_t *cue;
	dmxpair_t *p;
	/* check program id */
	if( (pid < 0) || (pid>=DMXPLAYER_PROGRAMS) ) return 0;
	/* append a new cue to the program */
	if(dmxprogram_append_new(programs[pid]) == 0) return 0;
	/* lock engine while writing */
	frametimer_lock();
		cue = dmxprogram_get_cue(programs[pid]->size-1, programs[pid]);
		p = programmer->first;
		while( p != 0)
		{
			dmxcue_set(p->channel, p->value, cue);
			p = dmxcue_get_next(p);
		}
	frametimer_unlock();
	return 1;
}



/**
 * dmxplayer_toogle_progam()
 * Toogle on/off a program
 * @param pid   Program id
 * @return 1 on success
**/
int dmxplayer_toogle_program(int pid)
{
	int n;
	/* check program id */
	if( (pid < 0) || (pid>=DMXPLAYER_PROGRAMS) ) return 0;
	/* only if program is INACTIVE and NON-EMPTY */
	if( (programs[pid]->active == 0) && (programs[pid]->size > 0) )
	{
		programs[pid]->pos     = 0;
		programs[pid]->active  = 1;
		programs[pid]->refresh = 1;
		if(programs[pid]->group != 0)
		{
			for(n=0; n<DMXPLAYER_PROGRAMS; n++)
				if( (programs[n]->group == programs[pid]->group) && (n!=pid) )
					programs[n]->active = 0;
		}
	}
	else programs[pid]->active = 0;
	return 1;
}



/**
 * dmxplayer_start_progam()
 * Starts a program
 * @param pid   Program id
 * @return 1 on success
**/
int dmxplayer_start_program(int pid)
{
	/* check program id */
	if( (pid < 0) || (pid>=DMXPLAYER_PROGRAMS) ) return 0;
	/* check if already running */
	if( (programs[pid]->active == 1) )
		return 1;
	/* use toogle function */
	else dmxplayer_toogle_program(pid);
	return 1;
}



/**
 * dmxplayer_stop_progam()
 * Stops a program
 * @param pid   Program id
 * @return 1 on success
**/
int dmxplayer_stop_program(int pid)
{
	/* check program id */
	if( (pid < 0) || (pid>=DMXPLAYER_PROGRAMS) ) return 0;
	/* use toogle function */
	programs[pid]->active = 0;
	return 1;
}



/**
 * dmxplayer_timesync()
**/
void dmxplayer_timesync()
{
	int pid;
	frametimer_lock();
		for(pid=0; pid<DMXPLAYER_PROGRAMS; pid++)
			programs[pid]->pos = 0;
		frametimer_sync();
	frametimer_unlock();
}



/**
 * dmxplayer_timetap()
**/
int dmxplayer_timetap()
{
	long begin =  get_time_usecs();
	long timediff = begin - lasttap;
	/* increase tap step */
	tapstep = (tapstep+1) % 4;
	/* if timediff > n seconds begin new tapping */
	if(timediff > (DMXPLAYER_TAPTIMEOUT*1000000)) 
	{
		tapstep = 0;
		frametimer_sync();
		timetap[tapstep] = steplen;
	}
	else
	{
		/* sync on first beat */
		if(tapstep == 0) frametimer_sync();
		/* set tap step time */
		timetap[tapstep] = timediff;
	}
	if(tapstep == 0) 
	{
		long sum = timetap[0] + timetap[1] + timetap[2] + timetap[3];
		/* no divide by 4, because tapping is in quarters */
		steplen = (sum * DMXPLAYER_FPS)/1000000;
	}
	/* give back tap step */
	lasttap = begin;
	return tapstep;
}

/**
 * dmxplayer_istaptimeout()
**/
int dmxplayer_istaptimeout()
{
	long begin =  get_time_usecs();
	long timediff = begin - lasttap;
	/* if timediff > 5 seconds begin new tapping */
	return (timediff > (DMXPLAYER_TAPTIMEOUT*1000000));
}



/**
 * dmxplayer_setbpm()
**/
void dmxplayer_setbpm(float bpm)
{
	if(bpm == 0.00f) return;
	steplen = (60.0f*4.0f) * (DMXPLAYER_FPS / (float)bpm);
}




/**
 * dmxplayer_getbpm()
 * Gets speed in beats per minute
**/
float dmxplayer_getbpm()
{
	return (60.0f*4.0f) * (DMXPLAYER_FPS / (float)steplen);
}




/**
 * dmxplayer_get_speed_sec()
 * Gets speed in seconds
**/
float dmxplayer_get_speed_sec()
{
	return 4.0f*(DMXPLAYER_FPS/(float)steplen);
}



/**
 * dmxplayer_getprogram()
 * Get a program by id
 * @param pid   Program index
 * @return Pointer to a dmxprogram_t
**/
dmxprogram_t *dmxplayer_getprogram(int pid)
{
	/* check program id */
	if( (pid < 0) || (pid>=DMXPLAYER_PROGRAMS) ) return 0;
	return programs[pid];
}



/**
 * Saves all programs into a file
 * @param fp    File pointer to store data in
 * @return Number of bytes written
**/
long dmxplayer_save(FILE *fp)
{
	int n;
	long len;
	len = 0;
	for(n=0; n<DMXPLAYER_PROGRAMS; n++)
	{
		len += dmxprogram_save(fp, programs[n]);
	}
	return len;
}



/**
 * Reads all programs from a file
 * @param fp   File pointer to read data from
 * @return Number of bytes read
**/
long dmxplayer_load(FILE *fp)
{
	int n;
	long len;
	len = 0;
	for(n=0; n<DMXPLAYER_PROGRAMS; n++)
	{
		len += dmxprogram_load(fp, programs[n]);
	}
	return len;
}



/**
 * Gets the quarter counter
**/
long dmxplayer_getQuarterCnt()
{
	return quarter_cnt;
}
