/**
 * @file main.c
 * @author Florian Schmitt <florian.schmitt@qikgrp.com>
 * ----------------------------------------------------------------------------
 * (c) Florian Schmitt / 2012-2013
 * License GPLv3+: GNU GPL Version 3 
 * or higher <http://gnu.org/licenses/gpl.html>
 * This code comes with ABSOLUTELY NO WARRANTY. This is free software, and 
 * you are welcome to redistribute it under certain conditions.
 * ----------------------------------------------------------------------------
 * A console program to make simple dmx setups
**/
#include "dmxgui.h"
#include "dmxset.h"
#include <stdio.h>
#include <signal.h>
#include <stdio_ext.h>
#include <stdlib.h>
#include <getopt.h>
#include "config.h"
#include "signals.h"
#include "midi/midirecv.h"

const char *program_name;


/**
 * print_usage()
 * Prints the program usage
 * @param stream     IOSTREAM to print to
 * @param exit_code  The code to give to exit() after printing
**/
void print_usage(FILE* stream, int exit_code)
{
	fprintf(stream, "Usage:  %s [options]\n", program_name);
	fprintf(stream, 
			"  -c  --channels n     Set the amount of channels\n"
			"  -h  --help           Display this usage information\n"
			"  -v  --version        Display version information\n"
			"  -p  --port n         Set the port to n\n"
			"  -l  --load file      Load a file on start-up\n"
			"  -s  --server n       Set the host to connect to\n");
	exit(exit_code);
}



/**
 * print_version()
 * Prints the program version
**/
void print_version()
{
	printf("%s: Version %d.%d.%d-%d\n", program_name, LINDMX_MAJOR_VERSION, 
	                                                   LINDMX_MINOR_VERSION, 
	                                                   LINDMX_PATCH_VERSION, 
	                                                   LINDMX_BUILD_NUMBER);
	printf("Copyright © 2013 Florian Schmitt\n"
	       "License GPLv3+: GNU GPL Version 3 or higher <http://gnu.org/licenses/gpl.html>\n"
	       "This program comes with ABSOLUTELY NO WARRANTY.\n"
	       "This is free software, and you are welcome to \n"
	       "redistribute it under certain conditions.\n\n"
	       "This software is written by Florian Schmitt <florian.schmitt@qikgrp.com>\n");
	exit(EXIT_SUCCESS);
}



/**
 * main()
 * Main procedure
**/
int main(int argc, char* argv[])
{
	int next_option, port, channels, ret;
	const char *remote = "localhost";
	const char *filename = 0;

	/* options */
	const char* const short_options = "hvs:p:c:l:";
	const struct option long_options[] = {
		{ "help",       0, NULL, 'h' },
		{ "version",    0, NULL, 'v' },
		{ "server",     1, NULL, 's' },
		{ "port",       1, NULL, 'p' },
		{ "channels",   1, NULL, 'c' },
		{ "load",       1, NULL, 'l' },
		{ NULL,       0, NULL,  0  }
	};

	/* default values */
	program_name = argv[0];
	port     = 6767;
	channels = 75;

	/* get the options */
	do {
		next_option = getopt_long(argc, argv, short_options, long_options, NULL);
		switch(next_option)
		{
			case 'h': 
				print_usage(stdout, EXIT_SUCCESS);
			case 'v': 
				print_version();
			case 'p': 
				port = atoi(optarg);
				break;
			case 'c': 
				channels = atoi(optarg);
				break;
			case 's': 
				remote = optarg;
				break;
			case 'l':
				filename = optarg;
				break;
			case '?': 
				print_usage(stderr, EXIT_FAILURE);
				break;
			case -1: 
				break;
			default:
				abort();
		}
	} while(next_option != -1);

	/* Set signal destination */
	signal(SIGINT, signals_handle_sig);

	ret = dmxset_global_init(remote, port);
	if(ret != DMXSET_ERR_NOERROR) 
	{
		fprintf(stderr, "Can not connect to lindmxd\n");
		exit(1);
	}
	midireceiver_init();

	dmxgui_run(channels, filename);

	midireceiver_destroy();
	dmxset_global_close();

	__fpurge(stdin);
	exit(EXIT_SUCCESS);
}
